import re
from enum import Enum
from enum import IntEnum

__all__ = (
    "SearchType",
    "TrackType",
    "PlaylistType",
    "NodeAlgorithm",
    "LoopMode",
    "RouteStrategy",
    "RouteIPType",
    "URLRegex",
    "LogLevel",
)


class SearchType(Enum):
    """
    The enum for the different search types for Pomice.
    This feature is exclusively for the Spotify search feature of Pomice.
    If you are not using this feature, this class is not necessary.

    SearchType.ytsearch searches using regular Youtube,
    which is best for all scenarios.

    SearchType.ytmsearch searches using YouTube Music,
    which is best for getting audio-only results.

    SearchType.scsearch searches using SoundCloud,
    which is an alternative to YouTube or YouTube Music.
    """

    ytsearch = "ytsearch"
    ytmsearch = "ytmsearch"
    scsearch = "scsearch"

    def __str__(self) -> str:
        return self.value


class TrackType(Enum):
    """
    The enum for the different track types for Pomice.

    TrackType.YOUTUBE defines that the track is from YouTube

    TrackType.SOUNDCLOUD defines that the track is from SoundCloud.

    TrackType.SPOTIFY defines that the track is from Spotify

    TrackType.APPLE_MUSIC defines that the track is from Apple Music.

    TrackType.HTTP defines that the track is from an HTTP source.

    TrackType.LOCAL defines that the track is from a local source.
    """

    # We don't have to define anything special for these, since these just serve as flags
    YOUTUBE = "youtube"
    SOUNDCLOUD = "soundcloud"
    SPOTIFY = "spotify"
    APPLE_MUSIC = "apple_music"
    HTTP = "http"
    LOCAL = "local"

    def __str__(self) -> str:
        return self.value


class PlaylistType(Enum):
    """
    The enum for the different playlist types for Pomice.

    PlaylistType.YOUTUBE defines that the playlist is from YouTube

    PlaylistType.SOUNDCLOUD defines that the playlist is from SoundCloud.

    PlaylistType.SPOTIFY defines that the playlist is from Spotify

    PlaylistType.APPLE_MUSIC defines that the playlist is from Apple Music.
    """

    # We don't have to define anything special for these, since these just serve as flags
    YOUTUBE = "youtube"
    SOUNDCLOUD = "soundcloud"
    SPOTIFY = "spotify"
    APPLE_MUSIC = "apple_music"

    def __str__(self) -> str:
        return self.value


class NodeAlgorithm(Enum):
    """
    The enum for the different node algorithms in Pomice.

    The enums in this class are to only differentiate different
    methods, since the actual method is handled in the
    get_best_node() method.

    NodeAlgorithm.by_ping returns a node based on it's latency,
    preferring a node with the lowest response time


    NodeAlgorithm.by_players return a nodes based on how many players it has.
    This algorithm prefers nodes with the least amount of players.
    """

    # We don't have to define anything special for these, since these just serve as flags
    by_ping = "BY_PING"
    by_players = "BY_PLAYERS"

    def __str__(self) -> str:
        return self.value


class LoopMode(Enum):
    """
    The enum for the different loop modes.
    This feature is exclusively for the queue utility of pomice.
    If you are not using this feature, this class is not necessary.

    LoopMode.TRACK sets the queue loop to the current track.

    LoopMode.QUEUE sets the queue loop to the whole queue.

    """

    # We don't have to define anything special for these, since these just serve as flags
    TRACK = "track"
    QUEUE = "queue"

    def __str__(self) -> str:
        return self.value


class RouteStrategy(Enum):
    """
    The enum for specifying the route planner strategy for Lavalink.
    This feature is exclusively for the RoutePlanner class.
    If you are not using this feature, this class is not necessary.

    RouteStrategy.ROTATE_ON_BAN specifies that the node is rotating IPs
    whenever they get banned by Youtube.

    RouteStrategy.LOAD_BALANCE specifies that the node is selecting
    random IPs to balance out requests between them.

    RouteStrategy.NANO_SWITCH specifies that the node is switching
    between IPs every CPU clock cycle.

    RouteStrategy.ROTATING_NANO_SWITCH specifies that the node is switching
    between IPs every CPU clock cycle and is rotating between IP blocks on
    ban.

    """

    ROTATE_ON_BAN = "RotatingIpRoutePlanner"
    LOAD_BALANCE = "BalancingIpRoutePlanner"
    NANO_SWITCH = "NanoIpRoutePlanner"
    ROTATING_NANO_SWITCH = "RotatingNanoIpRoutePlanner"


class RouteIPType(Enum):
    """
    The enum for specifying the route planner IP block type for Lavalink.
    This feature is exclusively for the RoutePlanner class.
    If you are not using this feature, this class is not necessary.

    RouteIPType.IPV4 specifies that the IP block type is IPV4

    RouteIPType.IPV6 specifies that the IP block type is IPV6
    """

    IPV4 = "Inet4Address"
    IPV6 = "Inet6Address"


class URLRegex:
    """
    The enum for all the URL Regexes in use by Pomice.

    URLRegex.SPOTIFY_URL returns the Spotify URL Regex.

    URLRegex.DISCORD_MP3_URL returns the Discord MP3 URL Regex.

    URLRegex.YOUTUBE_URL returns the Youtube URL Regex.

    URLRegex.YOUTUBE_PLAYLIST returns the Youtube Playlist Regex.

    URLRegex.YOUTUBE_TIMESTAMP returns the Youtube Timestamp Regex.

    URLRegex.AM_URL returns the Apple Music URL Regex.

    URLRegex.SOUNDCLOUD_URL returns the SoundCloud URL Regex.

    URLRegex.BASE_URL returns the standard URL Regex.

    """

    SPOTIFY_URL = re.compile(
        r"https?://open.spotify.com/(?P<type>album|playlist|track|artist)/(?P<id>[a-zA-Z0-9]+)",
    )

    DISCORD_MP3_URL = re.compile(
        r"https?://cdn.discordapp.com/attachments/(?P<channel_id>[0-9]+)/"
        r"(?P<message_id>[0-9]+)/(?P<file>[a-zA-Z0-9_.]+)+",
    )

    YOUTUBE_URL = re.compile(
        r"^((?:https?:)?\/\/)?((?:www|m)\.)?((?:youtube\.com|youtu.be))"
        r"(\/(?:[\w\-]+\?v=|embed\/|v\/)?)([\w\-]+)(\S+)?$",
    )

    YOUTUBE_PLAYLIST_URL = re.compile(
        r"^((?:https?:)?\/\/)?((?:www|m)\.)?((?:youtube\.com|youtu.be))/playlist\?list=.*",
    )

    YOUTUBE_TIMESTAMP = re.compile(
        r"(?P<video>^.*?)(\?t|&start)=(?P<time>\d+)?.*",
    )

    AM_URL = re.compile(
        r"https?://music.apple.com/(?P<country>[a-zA-Z]{2})/"
        r"(?P<type>album|playlist|song|artist)/(?P<name>.+)/(?P<id>[^?]+)",
    )

    AM_SINGLE_IN_ALBUM_REGEX = re.compile(
        r"https?://music.apple.com/(?P<country>[a-zA-Z]{2})/(?P<type>album|playlist|song|artist)/"
        r"(?P<name>.+)/(?P<id>.+)(\?i=)(?P<id2>.+)",
    )

    SOUNDCLOUD_URL = re.compile(
        r"((?:https?:)?\/\/)?((?:www|m)\.)?soundcloud.com\/.*/.*",
    )

    SOUNDCLOUD_PLAYLIST_URL = re.compile(
        r"^(https?:\/\/)?(www.)?(m\.)?soundcloud\.com\/.*/sets/.*",
    )

    SOUNDCLOUD_TRACK_IN_SET_URL = re.compile(
        r"^(https?:\/\/)?(www.)?(m\.)?soundcloud\.com/[a-zA-Z0-9-._]+/[a-zA-Z0-9-._]+(\?in)",
    )

    LAVALINK_SEARCH = re.compile(r"(?P<type>ytm?|sc)search:")

    BASE_URL = re.compile(r"https?://(?:www\.)?.+")


class LogLevel(IntEnum):
    """
    The enum for specifying the logging level within Pomice.
    This class serves as shorthand for logging.<level>
    This enum is exclusively for the logging feature in Pomice.
    If you are not using this feature, this class is not necessary.


    LogLevel.DEBUG sets the logging level to "debug".

    LogLevel.INFO sets the logging level to "info".

    LogLevel.WARN sets the logging level to "warn".

    LogLevel.ERROR sets the logging level to "error".

    LogLevel.CRITICAL sets the logging level to "CRITICAL".

    """

    DEBUG = 10
    INFO = 20
    WARN = 30
    ERROR = 40
    CRITICAL = 50



__all__ = (
    "SpotifyRequestException",
    "InvalidSpotifyURL",
)


class SpotifyRequestException(Exception):
    """An error occurred when making a request to the Spotify API"""

    pass


class InvalidSpotifyURL(Exception):
    """An invalid Spotify URL was passed"""

    pass



from __future__ import annotations

import logging
import re
import time
from base64 import b64encode
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import aiohttp
import orjson as json

from .exceptions import InvalidSpotifyURL
from .exceptions import SpotifyRequestException
from .objects import *

__all__ = ("Client",)


GRANT_URL = "https://accounts.spotify.com/api/token"
REQUEST_URL = "https://api.spotify.com/v1/{type}s/{id}"
SPOTIFY_URL_REGEX = re.compile(
    r"https?://open.spotify.com/(?P<type>album|playlist|track|artist)/(?P<id>[a-zA-Z0-9]+)",
)


class Client:
    """The base client for the Spotify module of Pomice.
    This class will do all the heavy lifting of getting all the metadata
    for any Spotify URL you throw at it.
    """

    def __init__(self, client_id: str, client_secret: str) -> None:
        self._client_id: str = client_id
        self._client_secret: str = client_secret

        self.session: aiohttp.ClientSession = None  # type: ignore

        self._bearer_token: Optional[str] = None
        self._expiry: float = 0.0
        self._auth_token = b64encode(
            f"{self._client_id}:{self._client_secret}".encode(),
        )
        self._grant_headers = {
            "Authorization": f"Basic {self._auth_token.decode()}",
        }
        self._bearer_headers: Optional[Dict] = None
        self._log = logging.getLogger(__name__)

    async def _set_session(self, session: aiohttp.ClientSession) -> None:
        self.session = session

    async def _fetch_bearer_token(self) -> None:
        _data = {"grant_type": "client_credentials"}

        resp = await self.session.post(GRANT_URL, data=_data, headers=self._grant_headers)

        if resp.status != 200:
            raise SpotifyRequestException(
                f"Error fetching bearer token: {resp.status} {resp.reason}",
            )

        data: dict = await resp.json(loads=json.loads)
        self._log.debug(f"Fetched Spotify bearer token successfully")

        self._bearer_token = data["access_token"]
        self._expiry = time.time() + (int(data["expires_in"]) - 10)
        self._bearer_headers = {
            "Authorization": f"Bearer {self._bearer_token}",
        }

    async def search(self, *, query: str) -> Union[Track, Album, Artist, Playlist]:
        if not self._bearer_token or time.time() >= self._expiry:
            await self._fetch_bearer_token()

        result = SPOTIFY_URL_REGEX.match(query)
        if not result:
            raise InvalidSpotifyURL("The Spotify link provided is not valid.")

        spotify_type = result.group("type")
        spotify_id = result.group("id")

        request_url = REQUEST_URL.format(type=spotify_type, id=spotify_id)

        resp = await self.session.get(request_url, headers=self._bearer_headers)
        if resp.status != 200:
            raise SpotifyRequestException(
                f"Error while fetching results: {resp.status} {resp.reason}",
            )

        data: dict = await resp.json(loads=json.loads)
        self._log.debug(
            f"Made request to Spotify API with status {resp.status} and response {data}",
        )

        if spotify_type == "track":
            return Track(data)
        elif spotify_type == "album":
            return Album(data)
        elif spotify_type == "artist":
            resp = await self.session.get(
                f"{request_url}/top-tracks?market=US",
                headers=self._bearer_headers,
            )
            if resp.status != 200:
                raise SpotifyRequestException(
                    f"Error while fetching results: {resp.status} {resp.reason}",
                )

            track_data: dict = await resp.json(loads=json.loads)
            tracks = track_data["tracks"]
            return Artist(data, tracks)
        else:
            tracks = [
                Track(track["track"])
                for track in data["tracks"]["items"]
                if track["track"] is not None
            ]

            if not len(tracks):
                raise SpotifyRequestException(
                    "This playlist is empty and therefore cannot be queued.",
                )

            next_page_url = data["tracks"]["next"]

            while next_page_url is not None:
                resp = await self.session.get(next_page_url, headers=self._bearer_headers)
                if resp.status != 200:
                    raise SpotifyRequestException(
                        f"Error while fetching results: {resp.status} {resp.reason}",
                    )

                next_data: dict = await resp.json(loads=json.loads)

                tracks += [
                    Track(track["track"])
                    for track in next_data["items"]
                    if track["track"] is not None
                ]
                next_page_url = next_data["next"]

            return Playlist(data, tracks)

    async def get_recommendations(self, *, query: str) -> List[Track]:
        if not self._bearer_token or time.time() >= self._expiry:
            await self._fetch_bearer_token()

        result = SPOTIFY_URL_REGEX.match(query)
        if not result:
            raise InvalidSpotifyURL("The Spotify link provided is not valid.")

        spotify_type = result.group("type")
        spotify_id = result.group("id")

        if not spotify_type == "track":
            raise InvalidSpotifyURL(
                "The provided query is not a Spotify track.",
            )

        request_url = REQUEST_URL.format(
            type="recommendation",
            id=f"?seed_tracks={spotify_id}",
        )

        resp = await self.session.get(request_url, headers=self._bearer_headers)
        if resp.status != 200:
            raise SpotifyRequestException(
                f"Error while fetching results: {resp.status} {resp.reason}",
            )

        data: dict = await resp.json(loads=json.loads)
        tracks = [Track(track) for track in data["tracks"]]

        return tracks



import random
import socket
import time
from datetime import datetime
from itertools import zip_longest
from timeit import default_timer as timer
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import NamedTuple
from typing import Optional

from .enums import RouteIPType
from .enums import RouteStrategy

__all__ = (
    "ExponentialBackoff",
    "NodeStats",
    "FailingIPBlock",
    "RouteStats",
    "Ping",
    "LavalinkVersion",
)


class ExponentialBackoff:
    """
    The MIT License (MIT)
    Copyright (c) 2015-present Rapptz
    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:
    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
    OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
    """

    def __init__(self, base: int = 1, *, integral: bool = False) -> None:
        self._base = base

        self._exp = 0
        self._max = 10
        self._reset_time = base * 2**11
        self._last_invocation = time.monotonic()

        rand = random.Random()
        rand.seed()

        self._randfunc = rand.randrange if integral else rand.uniform

    def delay(self) -> float:
        invocation = time.monotonic()
        interval = invocation - self._last_invocation
        self._last_invocation = invocation

        if interval > self._reset_time:
            self._exp = 0

        self._exp = min(self._exp + 1, self._max)
        return self._randfunc(0, self._base * 2**self._exp)  # type: ignore


class NodeStats:
    """The base class for the node stats object.
    Gives critical information on the node, which is updated every minute.
    """

    __slots__ = (
        "used",
        "free",
        "reservable",
        "allocated",
        "cpu_cores",
        "cpu_system_load",
        "cpu_process_load",
        "players_active",
        "players_total",
        "uptime",
    )

    def __init__(self, data: Dict[str, Any]) -> None:
        memory: dict = data.get("memory", {})
        self.used = memory.get("used")
        self.free = memory.get("free")
        self.reservable = memory.get("reservable")
        self.allocated = memory.get("allocated")

        cpu: dict = data.get("cpu", {})
        self.cpu_cores = cpu.get("cores")
        self.cpu_system_load = cpu.get("systemLoad")
        self.cpu_process_load = cpu.get("lavalinkLoad")

        self.players_active = data.get("playingPlayers")
        self.players_total = data.get("players")
        self.uptime = data.get("uptime")

    def __repr__(self) -> str:
        return f"<Pomice.NodeStats total_players={self.players_total!r} playing_active={self.players_active!r}>"


class FailingIPBlock:
    """
    The base class for the failing IP block object from the route planner stats.
    Gives critical information about any failing addresses on the block
    and the time they failed.
    """

    __slots__ = ("address", "failing_time")

    def __init__(self, data: dict) -> None:
        self.address = data.get("address")
        self.failing_time = datetime.fromtimestamp(
            float(data.get("failingTimestamp", 0)),
        )

    def __repr__(self) -> str:
        return f"<Pomice.FailingIPBlock address={self.address} failing_time={self.failing_time}>"


class RouteStats:
    """
    The base class for the route planner stats object.
    Gives critical information about the route planner strategy on the node.
    """

    __slots__ = (
        "strategy",
        "ip_block_type",
        "ip_block_size",
        "failing_addresses",
        "block_index",
        "address_index",
    )

    def __init__(self, data: Dict[str, Any]) -> None:
        self.strategy = RouteStrategy(data.get("class"))

        details: dict = data.get("details", {})

        ip_block: dict = details.get("ipBlock", {})
        self.ip_block_type = RouteIPType(ip_block.get("type"))
        self.ip_block_size = ip_block.get("size")
        self.failing_addresses = [
            FailingIPBlock(
                data,
            )
            for data in details.get("failingAddresses", [])
        ]

        self.block_index = details.get("blockIndex")
        self.address_index = details.get("currentAddressIndex")

    def __repr__(self) -> str:
        return f"<Pomice.RouteStats route_strategy={self.strategy!r} failing_addresses={len(self.failing_addresses)}>"


class Ping:
    # Thanks to https://github.com/zhengxiaowai/tcping for the nice ping impl
    def __init__(self, host: str, port: int, timeout: int = 5) -> None:
        self.timer = self.Timer()

        self._successed = 0
        self._failed = 0
        self._conn_time = None
        self._host = host
        self._port = port
        self._timeout = timeout

    class Socket:
        def __init__(self, family: int, type_: int, timeout: Optional[float]) -> None:
            s = socket.socket(family, type_)
            s.settimeout(timeout)
            self._s = s

        def connect(self, host: str, port: int) -> None:
            self._s.connect((host, port))

        def shutdown(self) -> None:
            self._s.shutdown(socket.SHUT_RD)

        def close(self) -> None:
            self._s.close()

    class Timer:
        def __init__(self) -> None:
            self._start: float = 0.0
            self._stop: float = 0.0

        def start(self) -> None:
            self._start = timer()

        def stop(self) -> None:
            self._stop = timer()

        def cost(self, funcs: Iterable[Callable], args: Any) -> float:
            self.start()
            for func, arg in zip_longest(funcs, args):
                if arg:
                    func(*arg)
                else:
                    func()

            self.stop()
            return self._stop - self._start

    def _create_socket(self, family: int, type_: int) -> Socket:
        return self.Socket(family, type_, self._timeout)

    def get_ping(self) -> float:
        s = self._create_socket(socket.AF_INET, socket.SOCK_STREAM)

        cost_time = self.timer.cost(
            (s.connect, s.shutdown),
            ((self._host, self._port), None),
        )
        s_runtime = 1000 * (cost_time)

        return s_runtime


class LavalinkVersion(NamedTuple):
    major: int
    minor: int
    fix: int

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LavalinkVersion):
            return False

        return (
            (self.major == other.major) and (self.minor == other.minor) and (self.fix == other.fix)
        )

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, LavalinkVersion):
            return False

        return not (self == other)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, LavalinkVersion):
            return False

        if self.major > other.major:
            return False
        if self.minor > other.minor:
            return False
        if self.fix > other.fix:
            return False
        return True

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, LavalinkVersion):
            return False

        return not (self < other)

    def __le__(self, other: object) -> bool:
        if not isinstance(other, LavalinkVersion):
            return False

        return (self < other) or (self == other)

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, LavalinkVersion):
            return False

        return (self > other) or (self == other)



"""Module for managing Apple Music objects"""
from typing import List

__all__ = (
    "Song",
    "Playlist",
    "Album",
    "Artist",
)


class Song:
    """The base class for an Apple Music song"""

    def __init__(self, data: dict) -> None:
        self.name: str = data["attributes"]["name"]
        self.url: str = data["attributes"]["url"]
        self.isrc: str = data["attributes"]["isrc"]
        self.length: float = data["attributes"]["durationInMillis"]
        self.id: str = data["id"]
        self.artists: str = data["attributes"]["artistName"]
        self.image: str = data["attributes"]["artwork"]["url"].replace(
            "{w}x{h}",
            f'{data["attributes"]["artwork"]["width"]}x{data["attributes"]["artwork"]["height"]}',
        )

    def __repr__(self) -> str:
        return (
            f"<Pomice.applemusic.Song name={self.name} artists={self.artists} "
            f"length={self.length} id={self.id} isrc={self.isrc}>"
        )


class Playlist:
    """The base class for an Apple Music playlist"""

    def __init__(self, data: dict, tracks: List[Song]) -> None:
        self.name: str = data["attributes"]["name"]
        self.owner: str = data["attributes"]["curatorName"]
        self.id: str = data["id"]
        self.tracks: List[Song] = tracks
        self.total_tracks: int = len(tracks)
        self.url: str = data["attributes"]["url"]
        # we'll use the first song's image as the image for the playlist
        # because apple dynamically generates playlist covers client-side
        self.image = self.tracks[0].image

    def __repr__(self) -> str:
        return (
            f"<Pomice.applemusic.Playlist name={self.name} owner={self.owner} id={self.id} "
            f"total_tracks={self.total_tracks} tracks={self.tracks}>"
        )


class Album:
    """The base class for an Apple Music album"""

    def __init__(self, data: dict) -> None:
        self.name: str = data["attributes"]["name"]
        self.url: str = data["attributes"]["url"]
        self.id: str = data["id"]
        self.artists: str = data["attributes"]["artistName"]
        self.total_tracks: int = data["attributes"]["trackCount"]
        self.tracks: List[Song] = [Song(track) for track in data["relationships"]["tracks"]["data"]]
        self.image: str = data["attributes"]["artwork"]["url"].replace(
            "{w}x{h}",
            f'{data["attributes"]["artwork"]["width"]}x{data["attributes"]["artwork"]["height"]}',
        )

    def __repr__(self) -> str:
        return (
            f"<Pomice.applemusic.Album name={self.name} artists={self.artists} id={self.id} "
            f"total_tracks={self.total_tracks} tracks={self.tracks}>"
        )


class Artist:
    """The base class for an Apple Music artist"""

    def __init__(self, data: dict, tracks: dict) -> None:
        self.name: str = f'Top tracks for {data["attributes"]["name"]}'
        self.url: str = data["attributes"]["url"]
        self.id: str = data["id"]
        self.genres: str = ", ".join(genre for genre in data["attributes"]["genreNames"])
        self.tracks: List[Song] = [Song(track) for track in tracks]
        self.image: str = data["attributes"]["artwork"]["url"].replace(
            "{w}x{h}",
            f'{data["attributes"]["artwork"]["width"]}x{data["attributes"]["artwork"]["height"]}',
        )

    def __repr__(self) -> str:
        return f"<Pomice.applemusic.Artist name={self.name} id={self.id} " f"tracks={self.tracks}>"



__all__ = (
    "AppleMusicRequestException",
    "InvalidAppleMusicURL",
)


class AppleMusicRequestException(Exception):
    """An error occurred when making a request to the Apple Music API"""

    pass


class InvalidAppleMusicURL(Exception):
    """An invalid Apple Music URL was passed"""

    pass
